package com.blazinggames.retro.classic.search;

// Java imports
import java.net.URL;

// Spelchan Imports
import com.blazinggames.urcl.data.Config;
import com.blazinggames.urcl.ui.*;
import com.blazinggames.urcl.text.*;

// Retro Imports
import com.blazinggames.retro.shared.search.tftm.*;

/**
*/
public class TftM 
extends TextDisplay
{
// *********************
// ***** CONSTANTS *****
// *********************
    
    private static final String VERSION = "1.10.040101";

    private static final String INSTRUCTIONS = "11INSTRUCTIONS";
	private static final String MAPNAMES = "11MAPNAME";
	private static final String MAPURL = "11MAPURL";
	
	private static final String COMMANDS[] = 
	{
		"?", "INFO", "I", "MAP", "L", "LEFT", "R", "RIGHT",
		"A", "ADVANCE", "FIRE", "END"
	};
	private static final int CMD_HELP = 0;
	private static final int CMD_INFO = 1;
	private static final int CMD_I = 2;
	private static final int CMD_MAP  = 3;
	private static final int CMD_L = 4;
	private static final int CMD_LEFT = 5;
	private static final int CMD_R = 6;
	private static final int CMD_RIGHT = 7;
	private static final int CMD_A = 8;
	private static final int CMD_ADVANCE = 9;
	private static final int CMD_FIRE = 10;
	private static final int CMD_END  = 11;
	private static final int NUM_COMMANDS = 12;

	private static final String FACINGS[] = 
	{
		"North", "North east", "East", "South east",
		"South", "South west", "West", "North west"
	};

// *********************
// ***** VARIABLES *****
// *********************

	private boolean _mapLoaded[];
	private String _mapNames[];
	private String _mapUrls[];
	private Map _mapData[];
    private Map _map;
    private URL _mapBase;

	private Instructions _instructions;
	
    // Configuration options
    private int _startingMoney = 10000;
    private int _tankCost = 5000;
    private int _factoryProfit = 250;
    private int _perTurnMoney = 100;

    // Player info
    private String _playerNames[] = new String[4];
    private boolean _playerHuman[] = new boolean[4];
    private int _playerMoney[] = new int[4];

	// AI
	private AIManager _ai[] = new AIManager[4];


// ************************
// ***** CONSTRUCTORs *****
// ************************

    public TftM(URL mapBase)
    {
        _mapBase = mapBase;
    }


// **************************************    
// ***** METHODS - From TextDisplay *****
// **************************************

    public boolean initialize(DisplayHandler dh)
    {
        super.initialize(dh);
        
        _map = new Map();
		_mapLoaded = new boolean[4];
		_mapNames = new String[4];
		_mapUrls = new String[4];
		_mapData = new Map[4];
        for (int cntr = 0; cntr < 4; ++cntr)
		{
			_mapLoaded[cntr] = false;
			_mapUrls[cntr] = _config.getStringProperty(MAPURL + "["+cntr+"]", "res/tftm/basic.map");
			_mapNames[cntr] = _config.getStringProperty(MAPNAMES + "[" + cntr + "]", "Unnamed");
			_mapData[cntr] = new Map();
		}
        return true;
    }
    

    public void run()
    {
        String s;
        int round;
        
        printHeader();
        do
        {
            initGame();
            commandShowMap();
            round = 1;
            do
            {
                println();
                println("*** Day " + round + " ***");
                ++round;
            }
            while (doTurn());
            
            if (_canQuit)
            {
                print("Do you want another game?");
                s = input();
                println(s);
                if ((s.startsWith("N")) || (s.startsWith("n")))
                    _displayHandler.displayEvent(this, "SHOW 1");
            }
        }
        while(_active);
        _displayHandler.log("TftMDisplay run thread exiting");
    }


// *************************************
// ***** METHODS - Game management *****
// *************************************

    private void printHeader()
    {
        println("*** Tanks for the Memory ***");
        println("Classic Version " + VERSION);
        println();
        println("Copyright (c) 2001-2002,2004");
		println("Blazing Games Inc. All Rights Reserved.");
        println();
    }

	private void instructions()
    {
		if (_instructions == null)
		{
			println("Grabbing instructions...");
			_res.addTextGroup(INSTRUCTIONS, _config.getStringProperty(INSTRUCTIONS, "tank.txt"));
			_instructions = new Instructions(_res.getTextGroup(INSTRUCTIONS));
			println("Done.");
			println();
		}
		_instructions.showInstructions(this);
    }

    private void initGame()
    {
        int cntr, aiNum = 1;
        int n = getRange("How many human players (1-4)? ", 1, 4);
        
        for (cntr = 0; cntr < n; ++cntr)
        {
            print("Player " + (cntr + 1) + " name:");
            _playerNames[cntr] = input();
            println(_playerNames[cntr]);
			_playerHuman[cntr] = true;
        }
        for (cntr = n; cntr < 4; ++cntr)
        {
            _playerNames[cntr] = "Computer " + aiNum;
            println("Player " + (cntr + 1) + " being named " + _playerNames[cntr]);
            ++aiNum;
			_playerHuman[cntr] = false;
        }
        
        initMap();
        for (cntr = n; cntr < 4; ++cntr)
        {
			_ai[cntr] = new AIManager(_map, cntr, _startingMoney, _perTurnMoney, _factoryProfit, _tankCost);
		}
		println();
		println("For Help type ? at command prompt");
		println();
    }

    private void initMap()
    {
		int cntr, choice;
		boolean validMap = false;
		do
		{
			println("Select Map");
			for (cntr = 0; cntr < 4; ++cntr)
			{
				println("  " + (cntr+1) + ". " +_mapNames[cntr]);
			}
			choice = getRange("Choice: ", 1, 4) - 1;
			if (_mapLoaded[choice] == false)
			{
				println("Loading map...");
				validMap = true;
				try
				{
					_mapData[choice].loadMap(new URL(_mapBase, _mapUrls[choice]));
					_mapLoaded[choice] = true;
				}
				catch (Exception e)
				{
					validMap = false;
					println("ERROR LOADING MAP");
					println(e.getMessage());
				}
				println("Done");
			}
			else
				validMap = true;
		}
		while (validMap == false);
		
		_map.cloneMap(_mapData[choice]);
		
		// standard game initializatoin
		for (cntr = 0; cntr < 4; ++cntr)
			_playerMoney[cntr] = _startingMoney;
    }
    
    private int getRange(String desc, int low, int high)
    {
        int result = high + 1;
        String s;
        
        do
        {
            print(desc);
            s = input();
            println(s);
            result = _config.parseInt(s, high + 1);
        }
        while ( (result < low) || (result > high) );
        
        return result;
    }

	/**
	 * Gets a particular character (case ignored)
	 *
	 * @param s the prompt string
	 * @param valid the list of valid UPPERCASE characters (in the form of a string)
	 * @return the index of the character that was entered
	*/
	private int getCharInput(String s, String valid)
	{
		String in, inUpper;
		int index, cntr, chValue;
		
		do
		{
			index = -1;
			print(s);
			in = input();
			println(in);
			inUpper = in.toUpperCase();
			if (inUpper.length() < 1)
				chValue = 0;
			else
				chValue = inUpper.charAt(0);
			for (cntr = 0; cntr < valid.length(); ++cntr)
			{
				if (valid.charAt(cntr) == chValue)
				{
					index = cntr;
					break;
				}
			}
		}
		while (index == -1);
		
		return index;
	}

// ******************************
// ***** METHODS - Gameflow *****
// ******************************

    /**
     * @return true if game is still in progress at end of turn
    */
    private boolean doTurn()
    {
		int cntr, temp;
		for (cntr = 0; cntr < 4; ++cntr)
		{
			println();
			println(_playerNames[cntr] + " Turn...");
			if (_playerHuman[cntr])
				humanTurn(cntr);
			else
				computerTurn(cntr);
		}
		
		temp = _map.endDay();
		if (temp < 0)
			return true;

		// This will only be reached when game over.
		println();
		String winMessage = "*** " +_playerNames[temp] + " has won ***";
		for (cntr = 0; cntr < winMessage.length(); ++cntr)
			print("*");
		println();
		println(winMessage);
		for (cntr = 0; cntr < winMessage.length(); ++cntr)
			print("*");
		println();
		println();
		return false;
    }

	private void humanTurn(int id)
	{
		int choice, temp, cntr, cntrCmd;
		String s, sUp;
		boolean turnInProgress;
		
		// new money
		println(" Got " + _perTurnMoney + " from funding.");
		_playerMoney[id] += _perTurnMoney;
		temp = _map._playerFactoryCounts[id];
		if (temp > 0)
		{
			cntr = temp * _factoryProfit;// don't want to create a new variable, so cheating by using cntr
			_playerMoney[id] += cntr;
			println(" Earned " + cntr + " from " + temp + " factories");
		}
		println (" You have " + _playerMoney[id]);
		
		if (_map.canCreateTank(id))
		{
			if (_playerMoney[id] >= _tankCost)
			{
				choice = getCharInput("Create new tank? ", "YN");
				if (choice == 0)
				{
					_map.createTank(id);
					_playerMoney[id] -= _tankCost;
				}
			}
			else
			{
				println("Not enough money for a new tank");
			}
		}
		else
			println("Base occupied so can not create tank");
		
		for (cntr = 0; cntr < 9; ++cntr)
		{
			if (_map._tanks[id][cntr].getHealth() > 0)
			{
				turnInProgress = (_map._tanks[id][cntr].startTurn() > 0);
				commandInfo(id, cntr);
				while (turnInProgress)
				{
					print ("Tank " + (cntr+1) +" command: ");
					choice = -1;
					s = input();
					println(s);
					for (cntrCmd = 0; cntrCmd < NUM_COMMANDS; ++cntrCmd)
					{
						if (s.equalsIgnoreCase(COMMANDS[cntrCmd]))
						{
							choice = cntrCmd;
							break;
						}
					}
					turnInProgress = processCommand(id, cntr, choice);
				}
			}
		}
	}
	
	private void computerTurn(int id)
	{
		String rs[] = _ai[id].playDay();
		
		for (int cntr = 0; cntr < rs.length; ++cntr)
			println(rs[cntr]);
	}
	
// *************************************
// ***** METHODS - Player Commands *****
// *************************************

	public boolean processCommand(int id, int tank, int cmd)
	{
		int choice, temp;
		boolean result = true;
		Tank t = _map._tanks[id][tank];
		 
		switch (cmd)
		{
			case CMD_HELP :
				instructions();
				break;

			case CMD_INFO :
			case CMD_I:
				commandInfo(id, tank);
				break;

			case CMD_MAP :
				commandShowMap();
				break;

			case CMD_L :
			case CMD_LEFT :
				t.turnLeft();
				println("Tank is now facing " + FACINGS[t.getFacing()]);
				if (_map.isTankOnBase(id, tank))
					t.useMovement(-10);
				println("Movement remaining: " + t.getMovement());
				break;

			case CMD_R :
			case CMD_RIGHT :
				t.turnRight();
				println("Tank is now facing " + FACINGS[t.getFacing()]);
				if (_map.isTankOnBase(id, tank))
					t.useMovement(-10);
				println("Movement remaining: " + t.getMovement());
				break;
				
			case CMD_A :
			case CMD_ADVANCE :
				temp = _map.moveTankForward(id, tank);
				if (temp < 0)
					println("You can not move there!");
				else if (temp > 0)
					println("You captured a factory!!!");
				println("Now at " + t.getX() + "," + t.getY());
				println("Movement remaining: " + t.getMovement());
				break;

			case CMD_FIRE :
				temp = _map.fire(id, tank);
				if (temp == -1)
				{
					println("You do not have ammo!");
				}
				else if (temp == 1)
				{
					println("You hit a tank");
				}
				else if (temp == 2)
				{
					println("You destroyed a tank!");
				}
				else
				{
					println("Shot never hit anything");
				}
				println("Movement remaining: " + t.getMovement());
				break;
				
			case CMD_END :
				result = false;
				break;

			default:
				println ("This command is unknown!");
				println ("Type ? for list of commands");
		}
		
		if (t.getMovement() < 1)
		{
			result = false;
		}
		
		return result;
	}

	private void commandInfo(int id, int tankID)
	{
		Tank tank = _map._tanks[id][tankID];
		int mx = Math.max(0, tank.getX() - 2);
		if (mx > 15) mx = 15;
		int my = Math.max(0, tank.getY() - 2);
		if (my > 15) my = 15;
		
		String s = _map.renderRow(my);
		println("  " + s.substring(mx, mx+5) + " TANK # " + (tankID + 1));
		s = _map.renderRow(my+1);
		println("  " + s.substring(mx, mx+5) + " loc: " + tank.getX() + "," + tank.getY() +
				" facing " + FACINGS[tank.getFacing()]);
		s = _map.renderRow(my+2);
		println("  " + s.substring(mx, mx+5) + " Health: " + tank.getHealth());
		s = _map.renderRow(my+3);
		println("  " + s.substring(mx, mx+5) + " Ammo remaning: " + tank.getAmmo());
		s = _map.renderRow(my+4);
		println("  " + s.substring(mx, mx+5) + " Movement: " + tank.getMovement());
	}
	
    private void commandShowMap()
    {
        println(" 01234567890123456789   KEY");
        for (int cntr = 0; cntr < 20; ++cntr)
        {
            print("" + (cntr % 10));
            print(_map.renderRow(cntr));
            print("   ");
            if (cntr < 16)
            {
                println(""+_map._tileRenderer.getSymbol(cntr)+" "+_map._tileRenderer.getName(cntr));
            }
            else
            {
                println("" + (cntr-15) + " Players tank");
            }
        }
    }
    
}
